"""TcEx Common Arg Handler"""
# standard library
import tempfile
from argparse import ArgumentParser, Namespace


class TcArgumentParser(ArgumentParser):
    """Overload of the ArgumentParser class.

    Adding common arguments for TcEx apps.
    """

    def __init__(self, **kwargs):
        """Initialize the Class properties."""
        super().__init__(**kwargs)
        # batch defaults
        self._batch_action = 'Create'
        self._batch_chunk = 25_000
        self._batch_halt_on_error = False
        self._batch_poll_interval = 15
        self._batch_poll_interval_max = 3_600
        self._batch_write_type = 'Append'

        # logger defaults
        self._tc_log_backup_count = 25  # target 50Mb in total log size (1x10Mb + 25x~1.6Mb = ~50Mb)
        self._tc_log_file = 'app.log'
        self._tc_log_level = 'info'
        self._tc_log_path = tempfile.gettempdir() or '/tmp'  # nosec
        self._tc_log_max_bytes = 10_485_760  # 10Mb
        self._tc_log_to_api = False

        # playbook defaults
        self._tc_playbook_db_type = 'Redis'
        self._tc_playbook_db_context = '1234-abcd'
        self._tc_playbook_db_path = 'localhost'
        self._tc_playbook_db_port = '6379'

        # service defaults
        self._tc_svc_hb_timeout_seconds = 20

        # standard defaults
        self._tc_api_path = 'https://api.threatconnect.com'
        self._tc_in_path = tempfile.gettempdir() or '/tmp'  # nosec
        self._tc_out_path = tempfile.gettempdir() or '/tmp'  # nosec
        self._tc_secure_params = False
        self._tc_temp_path = tempfile.gettempdir() or '/tmp'  # nosec
        self._tc_user_id = None

        # include arguments
        self._advanced_request()
        self._api_arguments()
        self._batch_arguments()
        self._cal_settings_arguments()
        self._logger_arguments()
        self._playbook_arguments()
        self._service_arguments()
        self._standard_arguments()

        # default namespace
        self.namespace = Namespace()

    def _advanced_request(self):
        """Define App Feature - Advanced Request args.

        --tc_adv_req_body params               The HTTP body.
        --tc_adv_req_exclude_null_params bool  If True, exclude null params.
        --tc_adv_req_fail_on_error bool        If True, fail on any encountered error.
        --tc_adv_req_headers params            The HTTP headers.
        --tc_adv_req_http_method method        The HTTP method (e.g., GET, POST).
        --tc_adv_req_params params             The HTTP query params.
        --tc_adv_req_path path                 The API path.
        --tc_adv_req_urlencode_body bool       If True, urlencode body.
        """
        self.add_argument('--tc_adv_req_body', default=None, help='The HTTP body.')
        self.add_argument(
            '--tc_adv_req_exclude_null_params',
            action='store_true',
            help='If True, exclude null params.',
        )
        self.add_argument(
            '--tc_adv_req_fail_on_error',
            action='store_true',
            default=None,
            help='If True, fail on any encountered error.',
        )
        self.add_argument('--tc_adv_req_headers', default=None, help='The HTTP headers.')
        self.add_argument(
            '--tc_adv_req_http_method', default=None, help='The HTTP method (e.g., GET, POST).'
        )
        self.add_argument('--tc_adv_req_params', default=None, help='The HTTP query params.')
        self.add_argument('--tc_adv_req_path', default=None, help='The API path.')
        self.add_argument(
            '--tc_adv_req_urlencode_body',
            action='store_true',
            default=None,
            help='If True, urlencode body.',
        )

    def _api_arguments(self):
        """Define TC API args.

        --tc_token token                  Token provided by ThreatConnect for app Authorization.
        --tc_token_expires token_expires  Expiration time for the passed Token.
        --api_access_id access_id         Access ID used for HMAC Authorization.
        --api_secret_key secret_key       Secret Key used for HMAC Authorization.
        """
        # TC main >= 4.4 token will be passed to jobs.
        self.add_argument('--tc_token', default=None, help='ThreatConnect API Token')
        self.add_argument(
            '--tc_token_expires',
            default=None,
            help='ThreatConnect API Token Expiration Time',
            type=int,
        )

        # TC Integrations Server or TC main < 4.4
        self.add_argument(
            '--api_access_id', default=None, help='ThreatConnect API Access ID', required=False
        )
        self.add_argument(
            '--api_secret_key', default=None, help='ThreatConnect API Secret Key', required=False
        )

        # Validate ThreatConnect SSL certificate
        self.add_argument(
            '--tc_verify', action='store_true', help='Validate the ThreatConnect SSL Cert'
        )

    def _batch_arguments(self):
        """Define Batch API args.

        --batch_action action          Action for the batch job ['Create', 'Delete'].
        --batch_chunk number           The maximum number of indicator per batch job.
        --batch_halt_on_error          Flag to indicate that the batch job should halt on error.
        --batch_poll_interval seconds  Seconds between batch status polls.
        --batch_interval_max seconds   Seconds before app should time out waiting on batch job
                                       completion.
        --batch_write_type type        Write type for Indicator attributes ['Append', 'Replace'].
        """
        self.add_argument(
            '--batch_action',
            choices=['Create', 'Delete'],
            default=self._batch_action,
            help='Action for the batch job',
        )
        self.add_argument(
            '--batch_chunk',
            default=self._batch_chunk,
            help='Max number of indicators per batch',
            type=int,
        )
        self.add_argument(
            '--batch_halt_on_error',
            action='store_true',
            default=self._batch_halt_on_error,
            help='Halt batch job on error',
        )
        self.add_argument(
            '--batch_poll_interval',
            default=self._batch_poll_interval,
            help='Frequency to run status check for batch job.',
            type=int,
        )
        self.add_argument(
            '--batch_poll_interval_max',
            default=self._batch_poll_interval_max,
            help='Maximum amount of time for status check on batch job.',
            type=int,
        )
        self.add_argument(
            '--batch_write_type',
            choices=['Append', 'Replace'],
            default=self._batch_write_type,
            help='Append or Replace attributes.',
        )

    def _cal_settings_arguments(self):
        """Define arguments TC sends for the CALSettings feature

        --tc_cal_host                  CAL host (just the hostname, without "https://").
        --tc_cal_token                 Authentication token for CAL.
        --tc_cal_timestamp             Timestamp to use with tc_cal_token to authenticate to CAL.
        """
        self.add_argument(
            '--tc_cal_host', help='CAL API host',
        )
        self.add_argument(
            '--tc_cal_token', help='CAL authentication token', type=int,
        )
        self.add_argument(
            '--tc_cal_timestamp', help='Timestamp to use with tc_cal_token to authenticate to CAL.',
        )

    def _logger_arguments(self):
        """Define logger args supported by every TcEx App.

        --tc_log_backup_count count  The number of backup files to keep.
        --tc_log_max_bytes bytes     The max size before rotating log file.

        --tc_log_file filename       The app log file name.
        --tc_log_path path           The app log path.
        --tc_log_to_api              Flag to indicate that app should log to API.
        --tc_log_level level         The logging level for the app.
        --logging level              Alias for **tc_log_level**.
        --log_curl                   Log external requests as CURL commands
        """
        self.add_argument(
            '--tc_log_backup_count',
            default=self._tc_log_backup_count,
            help='The number of backup files to keep',
        )
        self.add_argument(
            '--tc_log_max_bytes',
            default=self._tc_log_max_bytes,
            help='The max size before rotating log file',
        )
        self.add_argument('--tc_log_file', default=self._tc_log_file, help='App logfile name')
        self.add_argument('--tc_log_path', default=self._tc_log_path, help='ThreatConnect log path')
        # requires a token
        self.add_argument(
            '--tc_log_to_api',
            action='store_true',
            default=self._tc_log_to_api,
            help='ThreatConnect API Logging',
        )
        # self.add_argument(
        #     '--tc_log_level', '--logging', choices=['trace', 'debug', 'info', 'warning', 'error',
        #     'critical'],
        #     default=self._tc_log_level, help='Logging Level', dest='tc_log_level', type=str.lower)
        # BCS - temporarily until there is some way to configure App logging level in the UI
        self.add_argument(
            '--logging',
            choices=['trace', 'debug', 'info', 'warning', 'error', 'critical'],
            default=None,
            dest='logging',
            help='Logging Level',
            type=str.lower,
        )
        self.add_argument(
            '--tc_log_level',
            choices=['trace', 'debug', 'info', 'warning', 'error', 'critical'],
            default=None,
            dest='tc_log_level',
            help='Logging Level',
            type=str.lower,
        )
        self.add_argument(
            '--tc_log_curl',
            default=False,
            dest='tc_log_curl',
            help='Log external requests as CURL commands.',
            action='store_true',
        )

    def _playbook_arguments(self):
        """Define playbook specific args.

        These arguments will be passed to every playbook app by default.

        --tc_playbook_db_type type        The DB type (currently on Redis is supported).
        --tc_playbook_db_context context  The playbook context provided by TC.
        --tc_playbook_db_path path        The DB path or server name.
        --tc_playbook_db_port port        The DB port when required.
        --tc_playbook_out_variables vars  The output variable requested by downstream apps.
        """
        self.add_argument(
            '--tc_playbook_db_type', default=self._tc_playbook_db_type, help='Playbook DB type'
        )
        self.add_argument(
            '--tc_playbook_db_context',
            default=self._tc_playbook_db_context,
            help='Playbook DB Context',
        )
        self.add_argument(
            '--tc_playbook_db_path', default=self._tc_playbook_db_path, help='Playbook DB path'
        )
        self.add_argument(
            '--tc_playbook_db_port', default=self._tc_playbook_db_port, help='Playbook DB port'
        )
        self.add_argument(
            '--tc_playbook_out_variables', help='Playbook output variables', required=False
        )

    def _service_arguments(self):
        """Define service specific args.

        These arguments will be passed to every playbook app by default.

        --tc_svc_broker_cacert_file file            The Broker SSL CA (full chain) certificate.
        --tc_svc_broker_cert_file file              The Broker SSL Server certificate.
        --tc_svc_broker_conn_timeout seconds        The Broker service conn timeout (seconds).
        --tc_svc_broker_host host                   The Broker service hostname.
        --tc_svc_broker_jks_file file               Unused.
        --tc_svc_broker_jks_pwd password            Unused.
        --tc_svc_broker_port port                   The Broker service port.
        --tc_svc_broker_service service             The Broker service (mqtt/redis).
        --tc_svc_broker_timeout seconds             The Broker service timeout (tcex only).
        --tc_svc_broker_token token                 The Broker auth token.
        --tc_svc_client_topic topic                 The topic to send client message.
        --tc_svc_hb_timeout_seconds seconds         The heartbeat interval in seconds.
        --tc_svc_id id                              The unique service id.
        --tc_svc_server_topic topic                 The topic to receive server message.
        --tcex_testing_context context              A context for TcEx Framework testing.
        """
        self.add_argument('--tc_svc_broker_cacert_file', help='Broker client ssl certificate')
        self.add_argument('--tc_svc_broker_cert_file', help='Broker client ssl certificate')
        self.add_argument(
            '--tc_svc_broker_conn_timeout',
            default=10,
            help='Broker service conn timeout (seconds)',
            type=int,
        )
        self.add_argument('--tc_svc_broker_host', help='Broker service host')
        self.add_argument('--tc_svc_broker_jks_file', help='Unused')
        self.add_argument('--tc_svc_broker_jks_pwd', help='Unused')
        self.add_argument('--tc_svc_broker_port', help='Broker service port', type=int)
        self.add_argument(
            '--tc_svc_broker_service', default='mqtt', help='Broker service (mqtt/redis)'
        )
        self.add_argument(
            '--tc_svc_broker_timeout', default=60, help='Broker timeout in seconds', type=int
        )
        self.add_argument('--tc_svc_broker_token', help='Broker service auth token')
        self.add_argument('--tc_svc_client_topic', help='Topic to send client messages')
        self.add_argument(
            '--tc_svc_hb_timeout_seconds',
            default=self._tc_svc_hb_timeout_seconds,
            help='Broker heartbeat interval in seconds',
            type=int,
        )
        self.add_argument('--tc_svc_id', help='The unique service id')
        self.add_argument('--tc_svc_server_topic', help='Topic to send server messages')
        self.add_argument('--tcex_testing_context', help='A context for TcEx Framework testing')

    def _standard_arguments(self):
        """Define standard args passed to every TcEx App.

        --api_default_org org        The TC API user default organization.
        --tc_api_path path           The TC API path (e.g https://api.threatconnect.com).
        --tc_in_path path            The app in path.
        --tc_job_id id               The id of the running job.
        --tc_out_path path           The app out path.
        --tc_secure_params bool      Flag to indicator secure params is supported.
        --tc_temp_path path          The app temp path.
        --tc_user_id id              The user id of user running the job.

        --tc_proxy_host host         The proxy host.
        --tc_proxy_port port         The proxy port.
        --tc_proxy_username user     The proxy username.
        --tc_proxy_password pass     The proxy password.
        --tc_proxy_external          Flag to indicate external communications requires the use of a
                                     proxy.
        --tc_proxy_tc                Flag to indicate TC communications requires the use of a proxy.
        """
        self.add_argument('--api_default_org', default=None, help='ThreatConnect api default Org')
        self.add_argument(
            '--tc_action_channel', default=None, help='ThreatConnect AOT action channel'
        )
        self.add_argument('--tc_aot_enabled', action='store_true', help='ThreatConnect AOT enabled')
        self.add_argument('--tc_api_path', default=self._tc_api_path, help='ThreatConnect api path')
        self.add_argument('--tc_exit_channel', default=None, help='ThreatConnect AOT exit channel')
        self.add_argument('--tc_in_path', default=self._tc_in_path, help='ThreatConnect in path')
        self.add_argument('--tc_job_id', help='The id of the running job')
        self.add_argument(
            '--tc_out_path', default=self._tc_out_path, help='ThreatConnect output path'
        )
        self.add_argument(
            '--tc_secure_params',
            action='store_true',
            default=self._tc_secure_params,
            help='ThreatConnect Secure params enabled',
        )
        self.add_argument(
            '--tc_terminate_seconds',
            default=None,
            help='ThreatConnect AOT terminate seconds',
            type=int,
        )
        self.add_argument(
            '--tc_temp_path', default=self._tc_temp_path, help='ThreatConnect temp path'
        )
        self.add_argument('--tc_user_id', default=self._tc_user_id, help='User ID')

        # Proxy Configuration
        self.add_argument('--tc_proxy_host', default=None, help='Proxy Host')
        self.add_argument('--tc_proxy_port', default=None, help='Proxy Port')
        self.add_argument('--tc_proxy_username', default=None, help='Proxy User')
        self.add_argument('--tc_proxy_password', default=None, help='Proxy Password')
        self.add_argument(
            '--tc_proxy_external',
            '--apply_proxy_external',
            action='store_true',
            default=False,
            dest='tc_proxy_external',
            help='Proxy External Connections',
        )
        self.add_argument(
            '--tc_proxy_tc',
            '--apply_proxy_tc',
            action='store_true',
            default=False,
            dest='tc_proxy_tc',
            help='Proxy TC Connection',
        )

    def add_argument(self, *args, **kwargs):
        """Customize add_argument method for parser.

        Remove required flag from App args and instead insert a default value from default_args
        namespace. This is an attempt to move away from manipulating sys.argv and using args
        in favor on adding args in as a config (dict) in all cases.
        """
        if kwargs.get('required', False):
            # iterate over all args values to handle arg aliases
            for name in args:
                # strip '--' or '-' prefix from arg name.
                name = name.lstrip('-')
                # lookup default value in default_arg namespace
                if name in vars(self.namespace):
                    kwargs['default'] = getattr(self.namespace, name, None)
                    del kwargs['required']  # remove required fields to prevent argparser failure
                    break  # first match wins
        super().add_argument(*args, **kwargs)



"""Thread File Handler Class"""
# standard library
import os
import threading
from logging.handlers import RotatingFileHandler
from typing import Optional


class ThreadFileHandler(RotatingFileHandler):
    """Logger handler for ThreatConnect Exchange File logging."""

    handler_key = None
    thread_key = None

    def __init__(
        self,
        filename: str,
        mode: Optional[str] = 'a',
        maxBytes: Optional[int] = 0,
        backupCount: Optional[int] = 0,
        encoding: Optional[str] = None,
        delay: Optional[bool] = False,
    ):
        """Add logic to create log directory if it does not exists.

        Args:
            filename: The name of the logfile.
            mode: The write mode for the file.
            maxBytes: The max file size before rotating.
            backupCount: The maximum # of backup files.
            encoding: The log file encoding.
            delay: If True, then file opening is deferred until the first call to emit().
        """
        if encoding is None and os.getenv('LANG') is None:
            encoding = 'UTF-8'
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        RotatingFileHandler.__init__(self, filename, mode, maxBytes, backupCount, encoding, delay)

    def emit(self, record: object) -> None:
        """Emit a record.

        Emit logging events only if handler_key matches thread_key.

        Args:
            record: The record to be logged.
        """
        # handler_key and thread_key are added in logger.add_thread_file_handler() method
        if hasattr(threading.current_thread(), self.thread_key):
            if self.handler_key == getattr(threading.current_thread(), self.thread_key):
                RotatingFileHandler.emit(self, record)



"""ThreatConnect Task"""
from .api_endpoints import ApiEndpoints
from .assignee import Assignee
from .common_case_management import CommonCaseManagement
from .common_case_management_collection import CommonCaseManagementCollection
from .filter import Filter
from .tql import TQL


class Tasks(CommonCaseManagementCollection):
    """Tasks Class for Case Management Collection

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        initial_response (dict, optional): Initial data in
            Case Object for Task. Defaults to None.
        tql_filters (list, optional): List of TQL filters. Defaults to None.
        params(dict, optional): Dict of the params to be sent while
            retrieving the Task objects.
    """

    def __init__(self, tcex, initial_response=None, tql_filters=None, params=None):
        """Initialize Class properties."""
        super().__init__(
            tcex,
            ApiEndpoints.TASKS,
            initial_response=initial_response,
            tql_filters=tql_filters,
            params=params,
        )

        if initial_response:
            for item in initial_response.get('data', []):
                self.added_items.append(Task(tcex, **item))

    def __iter__(self):
        """Object iterator"""
        return self.iterate(initial_response=self.initial_response)

    def add_task(self, task):
        """Add an task to a case.

        Args:
            task (case_management.task.Task): The Task Object to add.
        """
        self.added_items.append(task)

    def entity_map(self, entity):
        """Map a dict to a Task.

        Args:
            entity (dict): The Task data.

        Returns:
            CaseManagement.Task: An Task Object
        """
        return Task(self.tcex, **entity)

    @property
    def filter(self):
        """Return instance of FilterTasks Object."""
        return FilterTasks(ApiEndpoints.TASKS, self.tcex, self.tql)


class Task(CommonCaseManagement):
    """Task object for Case Management.

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        artifacts (Artifact, kwargs): a list of Artifacts corresponding to the Task
        assignee (Assignee, kwargs): the user or group Assignee object for the Task
        case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Task.
        case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid** for the Task.
        completed_by (str, kwargs): [Read-Only] The **Completed By** for the Task.
        completed_date (str, kwargs): the completion date of the Task
        config_playbook (str, kwargs): [Read-Only] The **Config Playbook** for the Task.
        config_task (dict, kwargs): [Read-Only] The **Config Task** for the Task.
        dependent_on_id (str, kwargs): [Read-Only] The **Dependent On ID** for the Task.
        description (str, kwargs): The **Description** for the Task.
        due_date (str, kwargs): the due date of the Task
        duration (int, kwargs): [Read-Only] The **Duration** for the Task.
        id_dependent_on (int, kwargs): [Read-Only] The **Id_Dependent On** for the Task.
        name (str, kwargs): [Required] The **Name** for the Task.
        notes (Note, kwargs): a list of Notes corresponding to the Task
        parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Task.
        required (bool, kwargs): [Read-Only] The **Required** flag for the Task.
        status (str, kwargs): The **Status** for the Task.
        workflow_phase (int, kwargs): [Read-Only] the phase of the workflow
        workflow_step (int, kwargs): [Read-Only] the step of the workflow
        xid (str, kwargs): The **Xid** for the Task.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties"""
        super().__init__(tcex, ApiEndpoints.TASKS, kwargs)
        self.task_filter = [
            {
                'keyword': 'taskId',
                'operator': TQL.Operator.EQ,
                'value': self.id,
                'type': TQL.Type.INTEGER,
            }
        ]

        self._artifacts = kwargs.get('artifact', None)
        self._assignee = kwargs.get('assignee', None)
        self._case_id = kwargs.get('case_id', None)
        self._case_xid = kwargs.get('case_xid', None)
        self._completed_by = kwargs.get('completed_by', None)
        self._completed_date = kwargs.get('completed_date', None)
        self._config_playbook = kwargs.get('config_playbook', None)
        self._config_task = kwargs.get('config_task', None)
        self._dependent_on_id = kwargs.get('dependent_on_id', None)
        self._description = kwargs.get('description', None)
        self._due_date = kwargs.get('due_date', None)
        self._duration = kwargs.get('duration', None)
        self._name = kwargs.get('name', None)
        self._notes = kwargs.get('notes', None)
        self._parent_case = kwargs.get('parent_case', None)
        self._required = kwargs.get('required', None)
        self._status = kwargs.get('status', None)
        self._workflow_phase = kwargs.get('workflow_phase', None)
        self._workflow_step = kwargs.get('workflow_step', None)
        self._xid = kwargs.get('xid', None)

    def add_artifact(self, **kwargs):
        """Add a artifact to the task"""
        self.artifacts.add_artifact(self.tcex.cm.artifact(**kwargs))

    def add_note(self, **kwargs):
        """Add a note to the task"""
        self.notes.add_note(self.tcex.cm.note(**kwargs))

    @property
    def artifacts(self):
        """Return the **Artifacts** for the Task."""
        if self._artifacts is None or isinstance(self._artifacts, dict):
            artifacts = self._artifacts or {}
            self._artifacts = self.tcex.cm.artifacts(
                initial_response=artifacts, tql_filters=self.task_filter
            )
        return self._artifacts

    @property
    def as_entity(self):
        """Return the entity representation of the Task."""
        return {'type': 'Task', 'value': self.name, 'id': self.id}

    @property
    def assignee(self):
        """Return the **Assignee** for the Task."""
        if isinstance(self._assignee, dict):
            return self.tcex.cm.assignee(
                type=self._assignee.get('type', 'User'), **self._assignee.get('data')
            )
        return self._assignee

    @assignee.setter
    def assignee(self, assignee):
        """Set the **Assignee** for the Task."""
        if isinstance(assignee, dict):
            assignee = Assignee(type=assignee.get('type'), **assignee.get('data'))
        self._assignee = assignee

    @property
    def case_id(self):
        """Return the **Case ID** for the Task"""
        return self._case_id

    @case_id.setter
    def case_id(self, case_id):
        """Set the **Case ID** for the Task"""
        self._case_id = case_id

    @property
    def case_xid(self):
        """Return the **Case XID** for the Task"""
        return self._case_xid

    @case_xid.setter
    def case_xid(self, case_xid):
        """Set the **Case XID** for the Task"""
        self._case_xid = case_xid

    @property
    def completed_by(self):
        """Return the **Completed By** for the Task"""
        return self._completed_by

    @property
    def completed_date(self):
        """Return the **Completed Date** for the Task"""
        return self._completed_date

    @completed_date.setter
    def completed_date(self, completed_date):
        """Set the **Completed Date** for the Task"""
        self._completed_date = completed_date

    @property
    def config_playbook(self):
        """Return the **Config Playbook** for the Task"""
        return self._config_playbook

    @property
    def config_task(self):
        """Return the **Config Task** for the Task"""
        if self._config_task is None:
            self._config_task = ConfigTasks()
        elif isinstance(self._config_task, list):
            self._config_task = ConfigTasks(config_tasks=self._config_task)
        return self._config_task

    @property
    def dependent_on_id(self):
        """Return the **Depend On ID** for the Task"""
        return self._dependent_on_id

    @property
    def description(self):
        """Return the **Description** for the Task"""
        return self._description

    @description.setter
    def description(self, description):
        """Set the **Description** for the Task"""
        self._description = description

    @property
    def due_date(self):
        """Return the **Due Date** for the Task"""
        return self._due_date

    @due_date.setter
    def due_date(self, due_date):
        """Set the **Completed Date** for the Task"""
        self._due_date = due_date

    @property
    def duration(self):
        """Return the **Duration** for the Task"""
        return self._duration

    def entity_mapper(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): An entity dict used to update the Object.
        """
        new_case = Task(self.tcex, **entity)
        self.__dict__.update(new_case.__dict__)

    @property
    def name(self):
        """Return the **Name** for the Task"""
        return self._name

    @name.setter
    def name(self, name):
        """Set the **Name** for the Task"""
        self._name = name

    @property
    def notes(self):
        """Return the **Notes** for the Task"""
        if self._notes is None or isinstance(self._notes, dict):
            notes = self._notes or {}
            self._notes = self.tcex.cm.notes(initial_response=notes, tql_filters=self.task_filter)
        return self._notes

    @notes.setter
    def notes(self, notes):
        """Set the **Notes** for the Task"""
        if isinstance(notes, dict):
            self._notes = self.tcex.cm.notes(initial_response=notes, tql_filters=self.task_filter)
        self._notes = notes

    @property
    def parent_case(self):
        """Return the **Parent Case** for the Task."""
        if self._parent_case:
            return self.tcex.cm.case(**self._parent_case)
        return self._parent_case

    @property
    def required(self):
        """Return the **Required** for the Task"""
        return self._required

    @property
    def status(self):
        """Return the **Status** for the Task"""
        return self._status

    @status.setter
    def status(self, status):
        """Set the **Status** for the Task"""
        self._status = status

    @property
    def workflow_phase(self):
        """Return the **Workflow Phase** for the Task"""
        return self._workflow_phase

    @property
    def workflow_step(self):
        """Return the **Workflow Step** for the Task"""
        return self._workflow_step

    @property
    def xid(self):
        """Return the **XID** for the Task"""
        return self._xid

    @xid.setter
    def xid(self, xid):
        """Set the **XID** for the Task"""
        self._xid = xid


class ConfigTasks:
    """Sub class of the Cases object. Used to map the config tasks to.

    Args:
        config_tasks (list): A array of config tasks
    """

    def __init__(self, config_tasks=None):
        """Initialize Class properties."""
        if config_tasks is None:
            config_tasks = []
        self._config_tasks = []

        for config_task in config_tasks:
            if isinstance(config_task, ConfigTask):
                self._config_tasks.append(config_task)
            else:
                self._config_tasks.append(ConfigTask(**config_task))

    @property
    def config_tasks(self):
        """Return the **ConfigTasks**."""
        return self._config_tasks

    @property
    def as_dict(self):
        """Return a dict representation of the ConfigTask class."""
        data = []
        for task in self.config_tasks:
            data.append(task.as_dict)

        return data

    @property
    def body(self):
        """Return a body representation of the ConfigTask class."""
        return {}


class ConfigTask:
    """Config Task Object for Tasks"""

    def __init__(self, **kwargs):
        """Config Task object for Tasks."""
        self._transform_kwargs(kwargs)
        self._artifact_type = kwargs.get('artifact_type', None)
        self._data_type = kwargs.get('data_type', None)
        self._intel_type = kwargs.get('intel_type', None)
        self._name = kwargs.get('name', None)
        self._required = kwargs.get('required', None)
        self._ui_element = kwargs.get('ui_element', None)
        self._ui_label = kwargs.get('ui_label', None)

    @property
    def as_dict(self):
        """Return a dict representation of the Task Config class."""
        properties = vars(self)
        as_dict = {}
        for key, value in properties.items():
            key = key.lstrip('_')
            if value is None:
                continue
            as_dict[key] = value

        if not as_dict:
            return None

        return as_dict

    @property
    def artifact_type(self):
        """Return the **Artifact Type** for the Config Task"""
        return self._artifact_type

    @property
    def body(self):
        """Return a dict representation of the Task Config class."""
        return {}

    @property
    def data_type(self):
        """Return the **Data Type** for the Config Task"""
        return self._data_type

    @property
    def intel_type(self):
        """Return the **Intel Type** for the Config Task"""
        return self._intel_type

    @property
    def _metadata_map(self):
        """Return a mapping of kwargs to expected args."""
        return {
            'artifactType': 'artifact_type',
            'dataType': 'data_type',
            'intelType': 'intel_type',
            'uiElement': 'ui_element',
            'uiLabel': 'ui_label',
        }

    @property
    def name(self):
        """Return the **Name** for the Config Task"""
        return self._name

    @property
    def required(self):
        """Return the **Required** for the Config Task"""
        return self._required

    def _transform_kwargs(self, kwargs):
        """Map the provided kwargs to expected arguments."""
        for key in dict(kwargs):
            new_key = self._metadata_map.get(key, key)
            kwargs[new_key] = kwargs.pop(key)

    @property
    def ui_label(self):
        """Return the **UI Label** for the Config Task"""
        return self._ui_label

    @property
    def ui_element(self):
        """Return the **UI Element** for the Config Task"""
        return self._ui_element


class FilterTasks(Filter):
    """Filter Object for Tasks"""

    def automated(self, operator, automated):
        """Filter Tasks based on **automated** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            automated (bool): A flag indicating whether or not the task is automated.
        """
        self._tql.add_filter('automated', operator, automated, TQL.Type.BOOLEAN)

    def case_id(self, operator, case_id):
        """Filter Tasks based on **caseId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            case_id (int): The ID of the case this Task is associated with.
        """
        self._tql.add_filter('caseId', operator, case_id, TQL.Type.INTEGER)

    def case_severity(self, operator, case_severity):
        """Filter Tasks based on **caseSeverity** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            case_severity (str): The severity of the case associated with the task.
        """
        self._tql.add_filter('caseSeverity', operator, case_severity, TQL.Type.STRING)

    def completed_by(self, operator, completed_by):
        """Filter Tasks based on **completedBy** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            completed_by (str): The account login of the user who completed the task.
        """
        self._tql.add_filter('completedBy', operator, completed_by, TQL.Type.STRING)

    def completed_date(self, operator, completed_date):
        """Filter Tasks based on **completedDate** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            completed_date (str): The completion date for the task.
        """
        self._tql.add_filter('completedDate', operator, completed_date, TQL.Type.STRING)

    def description(self, operator, description):
        """Filter Tasks based on **description** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            description (str): The description of the task.
        """
        self._tql.add_filter('description', operator, description, TQL.Type.STRING)

    def due_date(self, operator, due_date):
        """Filter Tasks based on **dueDate** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            due_date (str): The due date for the task.
        """
        self._tql.add_filter('dueDate', operator, due_date, TQL.Type.STRING)

    # there is not way to add an **actual** artifact to a task through the API.
    @property
    def has_artifact(self):  # pragma: no cover
        """Return **FilterArtifacts** for further filtering."""
        from .artifact import FilterArtifacts

        artifacts = FilterArtifacts(ApiEndpoints.ARTIFACTS, self._tcex, TQL())
        self._tql.add_filter('hasArtifact', TQL.Operator.EQ, artifacts, TQL.Type.SUB_QUERY)
        return artifacts

    @property
    def has_case(self):
        """Return **FilterCases** for further filtering."""
        from .case import FilterCases

        cases = FilterCases(ApiEndpoints.CASES, self._tcex, TQL())
        self._tql.add_filter('hasCase', TQL.Operator.EQ, cases, TQL.Type.SUB_QUERY)
        return cases

    @property
    def has_note(self):
        """Return **FilterNotes** for further filtering."""
        from .note import FilterNotes

        notes = FilterNotes(ApiEndpoints.NOTES, self._tcex, TQL())
        self._tql.add_filter('hasNote', TQL.Operator.EQ, notes, TQL.Type.SUB_QUERY)
        return notes

    def id(self, operator, id):  # pylint: disable=redefined-builtin
        """Filter Tasks based on **id** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            id (int): The ID of the task.
        """
        self._tql.add_filter('id', operator, id, TQL.Type.INTEGER)

    def name(self, operator, name):
        """Filter Tasks based on **name** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            name (str): The name of the task.
        """
        self._tql.add_filter('name', operator, name, TQL.Type.STRING)

    def required(self, operator, required):
        """Filter Tasks based on **required** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            required (bool): Flag indicating whether or not the task is required.
        """
        self._tql.add_filter('required', operator, required, TQL.Type.BOOLEAN)

    def status(self, operator, status):
        """Filter Tasks based on **status** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            status (str): The status of the task.
        """
        self._tql.add_filter('status', operator, status, TQL.Type.STRING)

    def target_id(self, operator, target_id):
        """Filter Tasks based on **targetId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            target_id (int): The assigned user or group ID for the task.
        """
        self._tql.add_filter('targetId', operator, target_id, TQL.Type.INTEGER)

    def target_type(self, operator, target_type):
        """Filter Tasks based on **targetType** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            target_type (str): The target type for this task (either User or Group).
        """
        self._tql.add_filter('targetType', operator, target_type, TQL.Type.STRING)

    def workflow_phase(self, operator, workflow_phase):
        """Filter Tasks based on **workflowPhase** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            workflow_phase (int): The workflow phase of the task.
        """
        self._tql.add_filter('workflowPhase', operator, workflow_phase, TQL.Type.INTEGER)

    def workflow_step(self, operator, workflow_step):
        """Filter Tasks based on **workflowStep** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            workflow_step (int): The workflow step of the task.
        """
        self._tql.add_filter('workflowStep', operator, workflow_step, TQL.Type.INTEGER)

    def xid(self, operator, xid):
        """Filter Tasks based on **xid** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            xid (str): The XID of the task.
        """
        self._tql.add_filter('xid', operator, xid, TQL.Type.STRING)



"""ThreatConnect Threat Intelligence Filter Module"""


class Filters:
    """Filters module for Threat Intelligence"""

    def __init__(self, tcex):
        """Initialize class properties."""
        self._tcex = tcex
        self._filters = []

    @property
    def filters(self):
        """Return filters"""
        return self._filters

    def add_filter(self, filter_key, operator, value):
        """Add a filter given a key, operator, and value"""
        filter_key = self._metadata_map.get(filter_key, filter_key)
        self.filters.append({'name': filter_key, 'operator': operator, 'value': value})

    def remove_filter(self, filter_key):
        """Remove a filter, given a passed in filter key"""
        for f in list(self._filters):
            if f.get('name') == filter_key:
                self.filters.remove(f)

    @property
    def _metadata_map(self):
        """Return metadata map for Filter objects."""
        return {
            'date_added': 'dateAdded',
            'group.date_added': 'group.dateAdded',
            'document.file_type': 'document.fileType',
            'host.dns_active': 'host.dnsActive',
            'dns_active': 'dnsActive',
            'host.whois_active': 'host.whoisActive',
            'whois_active': 'whoisActive',
            'address.contry_name': 'address.contryName',
            'contry_name': 'contryName',
            'address.contry_code': 'address.contryCode',
            'contry_code': 'contryCode',
            'indicator.date_added': 'eventDate',
            'indicator.false_positive': 'false_positive',
            'false_positive': 'falsePositive',
            'indicator.threat_assess_score': 'indicator.threatAssessScore',
            'threat_assess_score': 'threatAssessScore',
            'indicator.threat_assess_confidence': 'indicator.threatAssessConfidence',
            'threat_assess_confidence': 'threatAssessConfidence',
            'indicator.threat_assess_rating': 'indicator.threatAssessRating',
            'threat_assess_rating': 'threatAssessRating',
        }

    @property
    def filters_string(self):
        """Return a filter string to be used as a param"""
        filter_strings = []
        for filter_key in self.filters:
            filter_strings.append(
                filter_key.get('name') + filter_key.get('operator') + filter_key.get('value')
            )

        return ','.join(filter_strings)



"""Service module for TcEx Framework"""
# flake8: noqa
from .api_service import ApiService
from .common_service_trigger import CommonServiceTrigger
from .mqtt_message_broker import MqttMessageBroker
from .webhook_trigger_service import WebhookTriggerService



"""TcEx Framework Webhook Service Trigger module."""
# standard library
import base64
import json
import traceback
from typing import Any, Callable, Optional, Union

from .common_service_trigger import CommonServiceTrigger


class WebhookTriggerService(CommonServiceTrigger):
    """TcEx Framework Webhook Service Trigger module."""

    def __init__(self, tcex: object):
        """Initialize the Class properties.

        Args:
            tcex: Instance of TcEx.
        """
        super().__init__(tcex)

        # config callbacks
        self.webhook_event_callback = None
        self.webhook_marshall_event_callback = None

    def callback_response_handler(self, callback_response: Any, message: dict) -> None:
        """Handle the different types of callback responses.

        # Webhook App (default)

        * Dict - Playbook will not be launched and provided data
                 will be used in the response to the client.
        * True - Playbook will be launched.
        * Else - Playbook will NOT be launched.

        # webhookResponseMarshall Feature App

        * Callable - Playbook will be launched and if marshall callback will be set to response.
        * True - Playbook will be launched.
        * Else - Playbook will NOT be launched.

        # webhookServiceEndpoint Feature App

        For this feature the callback method must fire the event on it's own.

        * Dict - Playbook will not be launched and provided data
                 will be used in the response to the client.
        * Else - Response will be set to default of statusCode=200, body=None, and headers=[].

        Args:
            callback_response: The response from the webhook callback method.
            message: The message payload from the server topic.
        """
        if self.ij.has_feature('webhookserviceendpoint'):
            self.callback_response_service_endpoint(callback_response, message)
        elif self.ij.has_feature('webhookresponsemarshall'):
            self.callback_response_marshall(callback_response, message)
        else:
            self.callback_response_webhook(callback_response, message)

    def callback_response_webhook(self, callback_response: Any, message: dict) -> None:
        """Handle the different types of callback responses.

        * Dict - Playbook will not be launched and provided data
                 will be used in the response to the client.
        * True - Playbook will be launched.
        * Else - Playbook will NOT be launched.

        Args:
            callback_response: The response from the webhook callback method.
            message: The message payload from the server topic.
        """
        if isinstance(callback_response, dict):
            # webhook responses are for providers that require a subscription req/resp.
            self.publish_webhook_event_response(message, callback_response)
        elif callback_response is True:
            self.increment_metric('Hits')
            self.fire_event_publish(
                message.get('triggerId'), self.session_id, message.get('requestKey')
            )

            # only required for testing in tcex framework
            self._tcex_testing(self.session_id, message.get('triggerId'))

            # capture fired status for testing framework
            self._tcex_testing_fired_events(self.session_id, True)
        else:
            self.increment_metric('Misses')

            # capture fired status for testing framework
            self._tcex_testing_fired_events(self.session_id, False)

    def callback_response_marshall(self, callback_response: Any, message: dict) -> None:
        """Handle the different types of callback responses.

        # webhookResponseMarshall Feature App

        * Callable - Playbook will be launched and if marshall callback will be set to response.
        * True - Playbook will be launched.
        * Else - Playbook will NOT be launched.

        Args:
            callback_response: The response from the webhook callback method.
            message: The message payload from the server topic.
        """
        fire_trigger = False
        if callable(callback_response):
            self.webhook_marshall_event_callback = callback_response
            fire_trigger = True

        # handle response the same a normal response
        self.callback_response_webhook(fire_trigger, message)

    def callback_response_service_endpoint(self, callback_response: Any, message: dict) -> None:
        """Handle the different types of callback responses.

        # webhookServiceEndpoint Feature App

        For this feature the callback method must fire the event on it's own.

        * Dict - Playbook will not be launched and provided data
                 will be used in the response to the client.
        * Else - Response will be set to default of statusCode=200, body=None, and headers=[].

        Args:
            callback_response: The response from the webhook callback method.
            message: The message payload from the server topic.
        """
        response = {
            'body': None,
            'headers': [],
            'statusCode': 200,
        }
        if isinstance(callback_response, dict):
            # webhook responses are for providers that require a subscription req/resp.
            response.update(callback_response)

        self.publish_webhook_event_response(message, callback_response)

    @property
    def command_map(self) -> dict:
        """Return the command map for the current Service type."""
        command_map: dict = super().command_map
        command_map.update({'webhookevent': self.process_webhook_event_command})
        command_map.update({'webhookmarshallevent': self.process_webhook_marshall_event_command})
        return command_map

    def process_webhook_event_command(self, message: dict) -> None:
        """Process the WebhookEvent command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
                "appId": 387,
                "command": "WebhookEvent",
                "triggerId": 1,
                "requestKey": "cd8c7a3a-7968-4b97-80c9-68b83a8ef1a1",
                "method": "GET",
                "headers": [
                    {
                        "name": "Accept-Encoding",
                        "value": "gzip, deflate, br"
                    }
                ],
                "queryParams": [
                    {
                        "name": "registration",
                        "value": "true"
                    }
                ]
            }

        Args:
            message: The message payload from the server topic.
        """
        self._create_logging_handler()
        self.log.trace(
            f'feature=webhook-trigger-service, event=process-webhook-event, message={message}'
        )

        # acknowledge webhook event (nothing is currently done with this message on the core side)
        self.publish_webhook_event_acknowledge(message)

        # get config using triggerId passed in WebhookEvent data
        config = None
        outputs = {}
        if not self.ij.has_feature('webhookserviceendpoint'):
            config: dict = self.configs.get(message.get('triggerId'))
            if config is None:
                self.log.error(
                    '''feature=webhook-trigger-service, event=missing-config, '''
                    f'''trigger-id={message.get('triggerId')}'''
                )
                return

            # get an instance of playbooks for App
            outputs: Union[list, str] = config.get('tc_playbook_out_variables') or []
            if isinstance(outputs, str):
                outputs = outputs.split(',')

        # get a context aware pb instance for the App callback method
        playbook: object = self.tcex.pb(context=self.session_id, output_variables=outputs)
        try:
            body: Any = self.redis_client.hget(message.get('requestKey'), 'request.body')
            if body is not None:
                body = base64.b64decode(body).decode()
            # pylint: disable=not-callable
            callback_data = {
                'body': body,
                'headers': message.get('headers'),
                'method': message.get('method'),
                'params': message.get('queryParams'),
            }
            if self.ij.has_feature('webhookresponsemarshall') or self.ij.has_feature(
                'webhookserviceendpoint'
            ):
                # add request_key arg when marshall or services endpoints feature is set (kwarg)
                callback_data.update({'request_key': message.get('requestKey')})
            elif not self.ij.has_feature('webhookserviceendpoint'):
                # add optional inputs for "standard" and "marshall" webhook trigger
                callback_data.update(
                    {
                        'config': config,
                        'playbook': playbook,
                        'trigger_id': message.get('triggerId'),
                    }
                )

            callback_response: Union[bool, Callable[..., Any], dict] = self.webhook_event_callback(
                **callback_data
            )
            self.callback_response_handler(callback_response, message)
        except Exception as e:
            self.increment_metric('Errors')
            self.log.error(
                'feature=webhook-trigger-service, event=webhook-callback-exception, '
                f'error="""{e}"""'
            )
            self.log.trace(traceback.format_exc())
        finally:
            self.logger.remove_handler_by_name(self.thread_name)

    def process_webhook_marshall_event_command(self, message: dict) -> None:
        """Process the WebhookMarshallEvent command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
                "appId": 95,
                "bodyVariable": "request.body",
                "command": "WebhookMarshallEvent",
                "headers": [
                    {
                        "name": "Accept",
                        "value": "*/*"
                    }
                ],
                "requestKey": "c29927c8-b94d-4116-a397-e6eb7002f41c",
                "statusCode": 200,
                "triggerId": 1234
            }

        Args:
            message: The message payload from the server topic.
        """
        self._create_logging_handler()
        self.log.trace(
            'feature=webhook-trigger-service, event=process-webhook-marshall-event, '
            f'message={message}'
        )

        # acknowledge webhook event (nothing is currently done with this message on the core side)
        self.publish_webhook_marshall_event_acknowledge(message)

        # get config using triggerId passed in WebhookMarshallEvent data
        config: dict = self.configs.get(message.get('triggerId'))
        if config is None:
            self.log.error(
                '''feature=webhook-trigger-service, event=missing-config, '''
                f'''trigger-id={message.get('triggerId')}'''
            )
            return

        body = None
        request_key: str = message.get('requestKey')

        try:
            body: Any = self.redis_client.hget(request_key, 'request.body')
            if body is not None:
                body = base64.b64decode(body).decode()
        except Exception as e:
            self.increment_metric('Errors')
            self.log.error(
                'feature=webhook-trigger-service, event=webhook-marshall-callback-exception, '
                f'error="""{e}"""'
            )
            self.log.trace(traceback.format_exc())

        # set default value for callback response to the data returned from the playbook
        response = {
            'body': body,
            'headers': message.get('headers'),
            'status_code': message.get('statusCode'),
        }

        if callable(self.webhook_marshall_event_callback):
            try:

                # call callback method
                # pylint: disable=not-callable
                callback_response: Optional[dict] = self.webhook_marshall_event_callback(
                    body=body,
                    headers=message.get('headers'),
                    request_key=request_key,
                    status_code=message.get('statusCode'),
                    trigger_id=message.get('triggerId'),
                )
                if isinstance(callback_response, dict):
                    response.update(callback_response)
            except Exception as e:
                self.increment_metric('Errors')
                self.log.error(
                    'feature=webhook-trigger-service, event=webhook-marshall-callback-exception, '
                    f'error="""{e}"""'
                )
                self.log.trace(traceback.format_exc())
            finally:
                self.logger.remove_handler_by_name(self.thread_name)

        # webhook responses are for providers that require a subscription req/resp.
        self.publish_webhook_event_response(message, response)

    def publish_webhook_event_acknowledge(self, message: dict) -> None:
        """Publish the WebhookEventResponse message.

        Args:
            message: The message from the broker.
        """
        self.message_broker.publish(
            json.dumps(
                {
                    'command': 'Acknowledged',
                    'requestKey': message.get('requestKey'),
                    'triggerId': message.get('triggerId'),
                    'type': 'WebhookEvent',
                }
            ),
            self.args.tc_svc_client_topic,
        )

    def publish_webhook_marshall_event_acknowledge(self, message: dict) -> None:
        """Publish the WebhookEventResponse message.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
                "command": "Acknowledged",
                "requestKey": "cd8c7a3a-7968-4b97-80c9-68b83a8ef1a1",
                "triggerId": 1,
                "type": "WebhookMarshallResponse"
            }

        Args:
            message: The message from the broker.
        """
        self.message_broker.publish(
            json.dumps(
                {
                    'command': 'Acknowledged',
                    'requestKey': message.get('requestKey'),
                    'triggerId': message.get('triggerId'),
                    'type': 'WebhookMarshallEvent',
                }
            ),
            self.args.tc_svc_client_topic,
        )

    def publish_webhook_event_response(self, message: dict, callback_response: dict) -> None:
        """Publish the WebhookEventResponse message.

        Args:
            message: The message from the broker.
            callback_response: The data from the callback method.
            playbook: Configure instance of Playbook used to write body.
        """
        playbook: object = self.tcex.pb(context=self.session_id, output_variables=[])

        # write response body to redis
        if callback_response.get('body') is not None:
            playbook.create_string(
                'response.body',
                base64.b64encode(callback_response.get('body').encode('utf-8')).decode('utf-8'),
            )

        # publish response
        self.message_broker.publish(
            json.dumps(
                {
                    'sessionId': self.session_id,  # session/context
                    'requestKey': message.get('requestKey'),
                    'command': 'WebhookEventResponse',
                    'triggerId': message.get('triggerId'),
                    'bodyVariable': 'response.body',
                    'headers': callback_response.get('headers', []),
                    'statusCode': callback_response.get('status_code', 200),
                }
            ),
            self.args.tc_svc_client_topic,
        )



"""ThreatConnect TI Group"""
# standard library
from urllib.parse import quote_plus

from ..mappings import Mappings


class Group(Mappings):
    """Unique API calls for Group API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties."""
        super().__init__(
            tcex,
            'Group',
            'groups',
            kwargs.pop('sub_type', None),
            kwargs.pop('api_entity', 'group'),
            kwargs.pop('api_branch', None),
            kwargs.pop('owner', None),
        )
        self.name = None

        for arg, value in kwargs.items():
            self.add_key_value(arg, value)

    @property
    def as_entity(self):
        """Return the entity representation of the Indicator."""
        return {
            'type': self.api_sub_type,
            'value': self.name,
            'id': int(self.unique_id) if self.unique_id else None,
        }

    @staticmethod
    def is_group():
        """Return True if object is a group."""
        return True

    @property
    def _metadata_map(self):
        """Return metadata map for Group objects."""
        return {
            'date_added': 'dateAdded',
            'event_date': 'eventDate',
            'file_name': 'fileName',
            'file_text': 'fileText',
            'file_type': 'fileType',
            'first_seen': 'firstSeen',
            'from_addr': 'from',
            'publish_date': 'publishDate',
            'to_addr': 'to',
        }

    def can_create(self):
        """Return True if group can be create."""
        return self.data.get('name') is not None

    def add_key_value(self, key, value):
        """Convert the value and adds it as a data field.

        Args:
            key:
            value:
        """
        key = self._metadata_map.get(key, key)
        if key in ['dateAdded', 'eventDate', 'firstSeen', 'publishDate']:
            self._data[key] = self._utils.datetime.format_datetime(
                value, date_format='%Y-%m-%dT%H:%M:%SZ'
            )
        elif key == 'file_content':
            # file content arg is not part of Group JSON
            pass
        elif key in ['unique_id', 'id']:
            self._unique_id = quote_plus(str(value))
        else:
            self._data[key] = value
        if key == 'name':
            self.name = value

    @property
    def name(self):
        """Return the Group name."""
        return self._data.get('name')

    @name.setter
    def name(self, name):
        """Set the Group name."""
        self._data['name'] = name

    def _set_unique_id(self, json_response):
        """Set the unique id of the Group."""
        self.unique_id = json_response.get('id', '')

    @staticmethod
    def sub_types():
        """All supported ThreatConnect Group types."""
        return {
            'Adversary': {'apiBranch': 'adversaries', 'apiEntity': 'adversary'},
            'Campaign': {'apiBranch': 'campaigns', 'apiEntity': 'campaign'},
            'Document': {'apiBranch': 'documents', 'apiEntity': 'document'},
            'Emails': {'apiBranch': 'emails', 'apiEntity': 'email'},
            'Event': {'apiBranch': 'events', 'apiEntity': 'event'},
            'Incident': {'apiBranch': 'incidents', 'apiEntity': 'incident'},
            'Intrusion Set': {'apiBranch': 'intrusionSets', 'apiEntity': 'intrusionSet'},
            'Report': {'apiBranch': 'reports', 'apiEntity': 'report'},
            'Signature': {'apiBranch': 'signatures', 'apiEntity': 'signature'},
            'Threat': {'apiBranch': 'threats', 'apiEntity': 'threat'},
        }



"""TcEx testing profile Class."""
# standard library
import json
import re
import sys
from typing import Optional

# third-party
import colorama as c

# autoreset colorama
c.init(autoreset=True, strip=False)


class Migrate:
    """Class for profile Migration methods management."""

    def __init__(self, profile: object):
        """Initialize Class properties."""
        self.profile = profile

    @property
    def _reserved_args(self) -> list:
        """Return a list of *all* ThreatConnect reserved arg values."""
        return [
            'api_access_id',
            'api_default_org',
            'api_secret_key',
            'batch_action',
            'batch_chunk',
            'batch_halt_on_error',
            'batch_interval_max',
            'batch_poll_interval',
            'batch_write_type',
            'logging',
            'tc_api_path',
            'tc_in_path',
            'tc_log_file',
            'tc_log_level',
            'tc_log_path',
            'tc_log_to_api',
            'tc_out_path',
            'tc_playbook_db_context',
            'tc_playbook_db_path',
            'tc_playbook_db_port',
            'tc_playbook_db_type',
            'tc_playbook_out_variables',
            'tc_proxy_host',
            'tc_proxy_port',
            'tc_proxy_username',
            'tc_proxy_password',
            'tc_proxy_external',
            'tc_proxy_tc',
            'tc_secure_params',
            'tc_temp_path',
            'tc_token',
            'tc_token_expires',
            'tc_user_id',
        ]

    def add_staging_data(self, profile_data: dict, name: str, type_: str, value: str) -> str:
        """Create staging data and return variable value.

        Args:
            profile_data: The profile to update with new staging data.
            name: The name of the input.
            type_: The type of input (Binary, StringArray, etc.)
            value: The value to write in the staging data.

        Returns:
            str: The variable for the staged data.
        """
        profile_data.setdefault('stage', {}).setdefault('kvstore', {})
        variable: str = self.profile.ij.create_variable(name, type_)
        profile_data['stage']['kvstore'][variable] = value
        return variable

    @staticmethod
    def comment_to_comments(profile_data: dict) -> None:
        """Migrate comment key to comments key.

        Args:
            profile_data: The profile data dict.
        """
        # APP-80 - migrate comment/comments field
        comment = profile_data.pop('comment', None)
        if isinstance(comment, str):
            profile_data['_comments_'] = [comment]
        elif isinstance(comment, list):
            profile_data['_comments_'] = comment

        comments = profile_data.pop('comments', None)
        if isinstance(comments, str):
            profile_data['_comments_'] = [comments]
        elif isinstance(comments, list):
            profile_data['_comments_'] = comments

    @property
    def data(self):
        """Migrate profile to latest schema and rewrite data.

        Called from test_case.py init_profile so that profile is update before
        the test case runs.
        """
        profile_data = self.profile.contents

        # migrate comment to comments and make an array
        self.comment_to_comments(profile_data)

        # remove any deprecated fields
        self.deprecated_fields(profile_data)

        # move default inputs
        self.move_default_inputs(profile_data)

        # update all env variables to match latest pattern
        self.permutation_output_variables(profile_data)

        # update config section of profile for service Apps
        self.service_config_inputs(profile_data)

        # create staged section if not already present
        self.create_staged_section(profile_data)

        # stage String inputs if they are not already staged
        self.stage_inputs(profile_data)

        # change for threatconnect staged data
        self.stage_redis_name(profile_data)

        # change for threatconnect staged data
        self.stage_threatconnect_data(profile_data)

        # update all version 1 env variables to match latest pattern
        profile_data = self.variable_pattern_env_v1(profile_data)

        # update all version 2 env variables to match latest pattern
        profile_data = self.variable_pattern_env_v2(profile_data)

        # update all tcenv variables to match latest pattern
        profile_data = self.variable_pattern_tcenv(profile_data)

        self.profile.write(profile_data)

    @staticmethod
    def deprecated_fields(profile_data: dict) -> None:
        """Remove deprecated fields.

        Args:
            profile_data: The profile data dict.
        """
        deprecated_fields = ['version']
        for d in deprecated_fields:
            try:
                del profile_data[d]
            except KeyError:
                pass

        # remove autostage
        deprecated_option_fields = ['autostage']
        for d in deprecated_option_fields:
            try:
                del profile_data['options'][d]
            except KeyError:
                pass

    def move_default_inputs(self, profile_data: dict) -> None:
        """Move any default args from optional or required inputs to defaults section.

        Args:
            profile_data: The profile data dict.
        """
        updated_params = []  # collect all PB configs for service Apps
        for profile_inputs in self.profile.profile_inputs:
            # ADI-1376 - handle tcex default args
            for input_type, inputs in dict(profile_inputs).items():
                for name in dict(inputs):
                    if name in self._reserved_args:
                        # preserve overwritten default arg
                        profile_inputs.setdefault('defaults', {})
                        profile_inputs['defaults'][name] = profile_inputs[input_type].pop(name)

            # updated params are for service Apps
            updated_params.append(profile_inputs)

        # use updated params from the validation section above to merge inputs
        if self.profile.ij.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']:
            for index, config_item in enumerate(profile_data.get('configs', [])):
                config_item['config'] = updated_params[index]
        else:
            # non-service App (PB Apps) will only have one set of inputs
            profile_data['inputs'] = updated_params[0]

    @staticmethod
    def permutation_output_variables(profile_data: dict) -> None:
        """Remove permutation_output_variables field.

        Args:
            profile_data: The profile data dict.
        """
        try:
            del profile_data['permutation_output_variables']
        except KeyError:
            pass

    def service_config_inputs(self, profile_data: dict) -> None:
        """Change flat config inputs to include required/options.

        Args:
            profile_data: The profile data dict.
        """
        for configs in profile_data.get('configs', []):
            config = configs.get('config', {})

            # handle updated configs
            if config.get('optional') or config.get('required'):
                continue

            # new config schema
            config_inputs = {'optional': {}, 'required': {}}

            # iterate over defined inputs
            for k, v in config.items():
                input_data = self.profile.ij.params_dict.get(k)
                input_type = 'optional'
                if input_data is not None:
                    if input_data.get('required') is True:
                        input_type = 'required'

                # add value back with appropriate input type
                config_inputs[input_type][k] = v

            # overwrite flattened config
            configs['config'] = config_inputs

    @staticmethod
    def create_staged_section(profile_data: dict) -> None:
        """Create the Stage and kvstore section in the profile if not currently present.

        Args:
            profile_data: The profile data dict.
        """
        if 'stage' not in profile_data.keys():
            profile_data['stage'] = {'kvstore': {}}
        if 'kvstore' not in profile_data.get('stage').keys():
            profile_data['stage']['kvstore'] = {}

    def stage_inputs(self, profile_data: dict) -> None:
        """Stage any non-staged profile data.

        Args:
            profile_data: The profile data dict.
        """
        if self.profile.ij.runtime_level.lower() != 'playbook':
            # staging is only required for PB Apps
            return

        for input_type in ['optional', 'required']:
            for k, v in dict(profile_data.get('inputs', {}).get(input_type, {})).items():
                # skip staging inputs with a null value
                if v is None:
                    variable: str = self.profile.ij.create_variable('null', 'String')
                    profile_data['inputs'][input_type][k] = variable
                    continue

                # get ij data for key/field
                ij_data = self.profile.ij.params_dict.get(k)

                # skip staging inputs that are not defined in install.json (possibly default args)
                if ij_data is None:
                    continue

                # skip staging for input types that don't support playbookDataType
                if ij_data.get('type').lower() in [
                    'boolean',
                    'choice',
                    'multichoice',
                ]:
                    continue

                # skip staging if variable if found in current input value
                if re.search(self.profile.utils.variable_parse, v):
                    continue

                # get the current PB data type, APP-607
                data_types = ij_data.get('playbookDataType', [])

                # only stage String values
                if 'String' not in data_types and not isinstance(v, str):
                    continue

                # convert input value to staged data
                v = self.add_staging_data(profile_data, k, 'String', v)

                # update input with new variable
                profile_data['inputs'][input_type][k] = v

    @staticmethod
    def stage_redis_name(profile_data: dict) -> None:
        """Update stage.redis to stage.kvstore

        This change updates the previous value of redis with a
        more generic value of kvstore for staged data.

        Args:
            profile_data: The current profile data dict.

        Returns:
            dict: The update profile dict.
        """
        if profile_data.get('stage') is None:
            return

        kvstore_data: Optional[dict] = profile_data['stage'].get('redis', None)
        if kvstore_data is not None:
            del profile_data['stage']['redis']
            if 'kvstore' not in profile_data['stage'].keys():
                profile_data['stage']['kvstore'] = kvstore_data

    @staticmethod
    def stage_threatconnect_data(profile_data: dict) -> None:
        """Update for staged threatconnect data section of profile

        This change updates the previous list to a dict with a key that
        can be reference as a variable in other sections of the profile.

        Args:
            profile_data: The current profile data dict.

        Returns:
            dict: The update profile dict.
        """
        if 'stage' not in profile_data:
            return

        stage_tc: Optional[dict] = profile_data.get('stage', {}).get('threatconnect')

        # check if profile is using old list type
        if isinstance(stage_tc, list):
            profile_data['stage']['threatconnect'] = {}

            counter = 0
            for item in stage_tc:
                profile_data['stage']['threatconnect'][f'item_{counter}'] = item
                counter += 1

    @staticmethod
    def variable_pattern_env_v1(profile_data: dict) -> dict:
        """Update the profile variable to latest pattern

        Args:
            profile_data: The profile data dict.

        Returns:
            dict: The updated dict.
        """
        profile = json.dumps(profile_data)

        for m in re.finditer(r'\"\$(env|envs)\.(\w+)\"', profile):
            full_match = m.group(0)
            try:
                env_type = m.group(1)
                env_key = m.group(2)

                new_variable = f'"${{{env_type}:{env_key}}}"'
                profile = profile.replace(full_match, new_variable)
            except IndexError:
                print(f'{c.Fore.YELLOW}Invalid variable found {full_match}.')
        return json.loads(profile)

    @staticmethod
    def variable_pattern_env_v2(profile_data: dict) -> dict:
        """Update the profile variable to latest pattern

        Args:
            profile_data: The profile data dict.

        Returns:
            dict: The updated dict.
        """
        profile = json.dumps(profile_data)

        for m in re.finditer(r'\${(env|envs|local|remote)\.(.*?)}', profile):
            full_match = m.group(0)
            try:
                env_type = m.group(1)  # currently env, envs, local, remote
                env_key = m.group(2)

                new_variable = f'${{{env_type}:{env_key}}}'
                profile = profile.replace(full_match, new_variable)
            except IndexError:
                print(f'{c.Fore.YELLOW}Invalid variable found {full_match}.')
        return json.loads(profile)

    def variable_pattern_tcenv(self, profile_data: dict) -> dict:
        """Update the profile variable to latest pattern

        Args:
            profile_data: The profile data dict.

        Returns:
            dict: The updated dict.
        """
        if 'jmespath' not in sys.modules:
            print(
                f'{c.Fore.RED}Missing jmespath module. Try '
                f'installing "pip install tcex[development]"'
            )
            sys.exit(1)

        profile = json.dumps(profile_data)

        for data in self.profile.tc_staged_data:
            key = data.get('key')
            for m in re.finditer(r'\${tcenv\.' + str(key) + r'\.(.*?)}', profile):
                full_match = m.group(0)
                try:
                    jmespath_expression = m.group(1)

                    new_variable = f'${{tcenv:{key}:{jmespath_expression}}}'
                    profile = profile.replace(full_match, new_variable)
                except IndexError:
                    print(f'{c.Fore.YELLOW}Invalid variable found {full_match}.')
        return json.loads(profile)



"""ThreatConnect TI Threat"""
from ..group import Group


class Threat(Group):
    """Unique API calls for Threat API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            name (str, kwargs): [Required for Create] The name for this Group.
            owner (str, kwargs): The name for this Group. Default to default Org when not provided
        """
        super().__init__(
            tcex, sub_type='Threat', api_entity='threat', api_branch='threats', **kwargs
        )



"""TcEx testing profile Class."""
# import json
# standard library
import json
import re

# third-party
import colorama as c

# autoreset colorama
c.init(autoreset=True, strip=False)


class Rules:
    """Class for profile Migration methods management."""

    def __init__(self, profile):
        """Initialize Class properties."""
        self.profile = profile

    def data(self, data: str) -> dict:  # pylint: disable=too-many-return-statements
        """Return the default output data for a given variable

        Args:
            data (any): The output data from the App for the current variable.

        Returns:
            []: [description]
        """

        # NOTE: The order of these if statements matter.
        if data is None:
            return {'expected_output': data, 'op': 'eq'}
        if isinstance(data, list) and not data:
            return {'expected_output': data, 'op': 'dd', 'ignore_order': False, 'exclude_paths': []}
        if self.matches_url_rule(data):
            return {'expected_output': data, 'op': 'is_url'}
        if self.matches_number_rule(data):
            return {'expected_output': data, 'op': 'is_number'}
        if self.matches_jeq_rule(data):
            return {
                'expected_output': data,
                'op': 'jeq',
                'ignore_order': False,
                'exclude_paths': [],
            }
        if self.matches_date_rule(data):
            return {'expected_output': data, 'op': 'is_date'}
        if self.matches_dd_rule(data):
            return {'expected_output': data, 'op': 'dd', 'ignore_order': False, 'exclude_paths': []}
        return {'expected_output': data, 'op': 'eq'}

    @staticmethod
    def matches_number_rule(outputs):
        """Return if output should use the is_number operator."""
        if not isinstance(outputs, list):
            outputs = [outputs]
        try:
            for output in outputs:
                int(output)
        except Exception:
            return False
        return True

    @staticmethod
    def matches_url_rule(outputs):
        """Return if output should use the is_url operator."""
        if not isinstance(outputs, list):
            outputs = [outputs]

        regex = re.compile(
            r'^(?:http|ftp)s?://'  # http:// or https://
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|[/?]\S+)$',
            re.IGNORECASE,
        )
        try:
            for output in outputs:
                matched = re.match(regex, output)
                if not matched:
                    return False
        except Exception:
            return False
        return True

    def matches_date_rule(self, outputs):
        """Return if output should use the is_date operator."""
        if not isinstance(outputs, list):
            outputs = [outputs]

        for output in outputs:
            # date_to_datatime will take int/floats and convert to date.
            # if the value can be converted to int/float then it will
            # not be considered a date for this rule.
            try:
                is_float = True
                float(output)
            except Exception:
                is_float = False
            else:
                if is_float:
                    return False

            try:
                is_int = True
                int(output)
            except Exception:
                is_int = False
            else:
                if is_int:
                    return False

            try:
                if self.profile.utils.datetime.date_to_datetime(output) is None:
                    return False
            except RuntimeError:
                return False

        return True

    @staticmethod
    def matches_jeq_rule(outputs):
        """Return if output should use the jeq operator."""
        # TODO: APP-674 - revisit this with Ben
        if not isinstance(outputs, list):
            outputs = [outputs]
        try:
            for output in outputs:
                if not isinstance(output, str):
                    return False
                if not isinstance(json.loads(output), (dict, list)):
                    return False
        except Exception:
            return False
        return True

    @staticmethod
    def matches_dd_rule(outputs):
        """Return if output should use the dd operator."""
        return isinstance(outputs, (dict, list))



"""ThreatConnect Case Management Artifact Type"""
from .api_endpoints import ApiEndpoints
from .common_case_management import CommonCaseManagement
from .common_case_management_collection import CommonCaseManagementCollection
from .filter import Filter
from .tql import TQL


class ArtifactTypes(CommonCaseManagementCollection):
    """Artifacts Type Class for Case Management Collection

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        initial_response (dict, optional): Initial data in
            Case Object for Artifact. Defaults to None.
        tql_filters (list, optional): List of TQL filters. Defaults to None.
        params(dict, optional): Dict of the params to be sent
            while retrieving the Artifact Type objects.
    """

    def __init__(self, tcex, initial_response=None, tql_filters=None, params=None):
        """Initialize Class properties."""
        super().__init__(
            tcex,
            ApiEndpoints.ARTIFACT_TYPES,
            initial_response=initial_response,
            tql_filters=tql_filters,
            params=params,
        )

    def __iter__(self):
        """Object iterator"""
        return self.iterate(initial_response=self.initial_response)

    def entity_map(self, entity):
        """Map a dict to a Artifact.

        Args:
            entity (dict): The Artifact data.

        Returns:
            CaseManagement.Artifact: An Artifact Object
        """
        return ArtifactType(self.tcex, **entity)

    @property
    def filter(self):
        """Return instance of FilterArtifactTypes Object."""
        return FilterArtifactTypes(ApiEndpoints.ARTIFACT_TYPES, self.tcex, self.tql)


class ArtifactType(CommonCaseManagement):
    """ArtifactType object for Case Management.

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        data_type (str, kwargs): [Read-Only] The **Data Type** for the Artifact Type.
        description (str, kwargs): [Read-Only] The **Description** for the Artifact Type.
        intel_type (str, kwargs): [Read-Only] The **Intel Type** for the Artifact Type.
        name (str, kwargs): [Read-Only] The **Name** for the Artifact Type.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties"""
        super().__init__(tcex, ApiEndpoints.ARTIFACT_TYPES, kwargs)
        self._data_type = kwargs.get('data_type', None)
        self._description = kwargs.get('description', None)
        self._intel_type = kwargs.get('intel_type', None)
        self._name = kwargs.get('name', None)

    @property
    def as_entity(self):
        """Return the entity representation of the Artifact Type."""
        return {'type': 'Artifact Type', 'value': self.name, 'id': self.id}

    def entity_mapper(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): An entity dict used to update the Object.
        """
        new_case = ArtifactType(self.tcex, **entity)
        self.__dict__.update(new_case.__dict__)

    @property
    def data_type(self):
        """Return the **Date Type** for the Artifact."""
        return self._data_type

    @property
    def description(self):
        """Return the **Description** for the Artifact."""
        return self._description

    @property
    def intel_type(self):
        """Return the **Intel Type** for the Artifact."""
        return self._intel_type

    @property
    def name(self):
        """Return the **Name** for the Artifact."""
        return self._name


class FilterArtifactTypes(Filter):
    """Filter Object for ArtifactTypes"""

    def active(self, operator, active):
        """Filter Artifact Types based on **active** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            active (bool): The active status of the artifact type.
        """
        self._tql.add_filter('active', operator, active, TQL.Type.BOOLEAN)

    def data_type(self, operator, data_type):
        """Filter Artifact Types based on **dataType** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            data_type (str): The data type of the artifact type.
        """
        self._tql.add_filter('dataType', operator, data_type, TQL.Type.STRING)

    def description(self, operator, description):
        """Filter Artifact Types based on **description** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            description (str): The description of the artifact type.
        """
        self._tql.add_filter('description', operator, description, TQL.Type.STRING)

    def id(self, operator, id):  # pylint: disable=redefined-builtin
        """Filter Artifact Types based on **id** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            id (int): The ID of the artifact type.
        """
        self._tql.add_filter('id', operator, id, TQL.Type.INTEGER)

    def intel_type(self, operator, intel_type):
        """Filter Artifact Types based on **intelType** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            intel_type (str): The intel type of the artifact type.
        """
        self._tql.add_filter('intelType', operator, intel_type, TQL.Type.STRING)

    def managed(self, operator, managed):
        """Filter Artifact Types based on **managed** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            managed (bool): The managed status of the artifact type.
        """
        self._tql.add_filter('managed', operator, managed, TQL.Type.BOOLEAN)

    def name(self, operator, name):
        """Filter Artifact Types based on **name** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            name (str): The name of the artifact type.
        """
        self._tql.add_filter('name', operator, name, TQL.Type.STRING)



"""TcEx testing Framework."""
# standard library
import json
import os
import random
import string
import time
import traceback
import uuid
from datetime import datetime

# third-party
import pytest
import urllib3
from requests import Session

# first-party
from tcex import TcEx
from tcex.app_config_object.install_json import InstallJson
from tcex.env_store import EnvStore
from tcex.profile import Profile
from tcex.utils import Utils

from .stage_data import Stager
from .test_logger import logger
from .test_session import TestSession
from .validate_data import Validator

# disable ssl warning message
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class TestCase:
    """Base TestCase Class"""

    _app_path = os.getcwd()
    _current_test = None
    _default_args = None
    _profile = None
    _session_admin = None
    _session_exchange = None
    _stager = None
    _staged_tc_data = []
    _timer_class_start = None
    _timer_method_start = None
    _validator = None
    app = None
    context = None
    enable_update_profile = False
    env = set(os.getenv('TCEX_TEST_ENVS', 'build').split(','))
    env_store = EnvStore(logger=logger)
    ij = InstallJson(logger=logger)
    initialized = False
    log = logger
    redis_client = None
    session = Session()
    tcex = None
    tcex_testing_context = None
    use_token = True
    utils = Utils()

    def _reset_property_flags(self):
        """Reset all control flag."""
        # used to prevent pytest from executing @property methods
        self.initialized = False
        self.tcex = False

    def _log_args(self, args):
        """Log args masking any that are marked encrypted and log warning for unknown args.

        Args:
            args (dict): A dictionary of args.
        """
        for name, value in sorted(args.items()):
            input_data = self.ij.params_dict.get(name)
            if input_data is None and name not in self.default_args:
                # if name is not in install.json and not a default arg log a warning
                if name not in ['tcex_testing_context']:  # values that should not cause a warning
                    self.log.data('run', 'input', f'{name}: unknown arg.', 'warning')
            elif input_data is not None and input_data.get('encrypt') is True:
                self.log.data('run', 'input', f'{name}: ***')
            else:
                self.log.data('run', 'input', f'{name}: {value}')

    @staticmethod
    def _to_bool(value):
        """Return bool value from int or string."""
        return str(value).lower() in ['1', 'true']

    def _update_args(self, args: dict) -> dict:
        """Update args before running App.

        Args:
            args: The current argument dictionary.

        Returns:
            dict: The updated argument dictionary.
        """

        if self.ij.runtime_level.lower() in ['playbook']:
            # set requested output variables
            args['tc_playbook_out_variables'] = self.profile.tc_playbook_out_variables

        # update path args
        self._update_path_args(args)

        # update default args with app args
        app_args = self.default_args.copy()
        app_args.update(args)

        # update default args with app args
        app_args['tc_logger_name'] = self.context

        # safely log all args to tests.log
        self._log_args(app_args)

        return app_args

    def _update_path_args(self, args):
        """Update path in args for each test profile."""
        # service Apps do not have a profile when this is needed.
        profile = self.profile or Profile(default_args=self.default_args.copy())
        args['tc_in_path'] = profile.tc_in_path
        args['tc_log_path'] = profile.tc_log_path
        args['tc_out_path'] = profile.tc_out_path
        args['tc_temp_path'] = profile.tc_temp_path

    @property
    def api_access_id(self):
        """Return TC API Access ID"""
        return self.env_store.getenv('/ninja/tc/tci/exchange_admin/api_access_id')

    @property
    def api_secret_key(self):
        """Return TC API Secret Key"""
        return self.env_store.getenv('/ninja/tc/tci/exchange_admin/api_secret_key')

    @staticmethod
    def check_environment(environments, os_environments=None):
        """Check if test case matches current environments, else skip test.

        Args:
            environments (list): The test case environments.
            os_environments (list): The user/os defined environment.
        """
        test_envs = environments or ['build']
        os_envs = set(os.environ.get('TCEX_TEST_ENVS', 'build').split(','))
        if os_environments:
            os_envs = set(os_environments)
        if not os_envs.intersection(set(test_envs)):
            pytest.skip('Profile skipped based on current environment.')

    def create_config(self, args):
        """Create files necessary to start a Service App."""
        config = self._update_args(args)

        # service Apps will get their args/params from encrypted file in the "in" directory
        data = json.dumps(config, sort_keys=True).encode('utf-8')
        key = ''.join(random.choice(string.ascii_lowercase) for i in range(16))
        encrypted_data = self.utils.encrypt_aes_cbc(key, data)

        # ensure that the in directory exists
        os.makedirs(config.get('tc_in_path'), exist_ok=True)

        # write the file in/.app_params.json
        app_params_json = os.path.join(config.get('tc_in_path'), '.test_app_params.json')
        with open(app_params_json, 'wb') as fh:
            fh.write(encrypted_data)

        # create environment variable for tcex inputs method to pick up to read encrypted file
        os.environ['TC_APP_PARAM_KEY'] = key
        os.environ['TC_APP_PARAM_FILE'] = app_params_json

    @property
    def default_args(self):
        """Return App default args."""
        if self._default_args is None and self.initialized:
            self._default_args = {
                # local override TCI_EXCHANGE_ADMIN_API_ACCESS_ID
                'api_access_id': self.api_access_id,
                'api_default_org': os.getenv('API_DEFAULT_ORG', 'TCI'),
                # local override TCI_EXCHANGE_ADMIN_API_SECRET_KEY
                'api_secret_key': self.api_secret_key,
                'tc_api_path': self.tc_api_path,
                'tc_in_path': os.getenv('TC_IN_PATH', 'log'),
                'tc_log_level': os.getenv('TC_LOG_LEVEL', 'trace'),
                'tc_log_path': os.getenv('TC_LOG_PATH', 'log'),
                'tc_log_to_api': self._to_bool(os.getenv('TC_LOG_TO_API', 'false')),
                'tc_logger_name': 'tcex',
                'tc_out_path': os.getenv('TC_OUT_PATH', 'log'),
                'tc_proxy_external': self._to_bool(os.getenv('TC_PROXY_EXTERNAL', 'false')),
                # local override TC_PROXY_HOST
                'tc_proxy_host': self.env_store.getenv(
                    '/ninja/proxy/tc_proxy_host', default='localhost'
                ),
                # local override TC_PROXY_PASSWORD
                'tc_proxy_password': self.env_store.getenv(
                    '/ninja/proxy/tc_proxy_password', default=''
                ),
                # local override TC_PROXY_PORT
                'tc_proxy_port': self.env_store.getenv(
                    '/ninja/proxy/tc_proxy_port', default='4242'
                ),
                'tc_proxy_tc': self._to_bool(os.getenv('TC_PROXY_TC', 'false')),
                # local override TC_PROXY_USERNAME
                'tc_proxy_username': self.env_store.getenv(
                    '/ninja/proxy/tc_proxy_username', default=''
                ),
                'tc_temp_path': os.getenv('TC_TEMP_PATH', 'log'),
            }

            # try to use token when possible
            if self.use_token is True:
                token = os.getenv('TC_TOKEN', self.tc_token)
                if token is not None:
                    # if token was successfully retrieved from TC use token and remove hmac values
                    self._default_args['tc_token'] = token
                    self._default_args['tc_token_expires'] = int(time.time()) + 3600
                    del self._default_args['api_access_id']
                    del self._default_args['api_secret_key']
        return self._default_args

    def init_profile(self, profile_name, pytestconfig=None, monkeypatch=None, options=None):
        """Stages and sets up the profile given a profile name"""
        self._profile = Profile(
            default_args=self.default_args.copy(),
            name=profile_name,
            pytestconfig=pytestconfig,
            monkeypatch=monkeypatch,
            redis_client=self.redis_client,
            tcex_testing_context=self.tcex_testing_context,
            logger=self.log,
            options=options,
        )

        # Override test environments if specified
        os_environments = None
        if pytestconfig:
            os_environments = pytestconfig.option.environment

        # check profile environment
        self.check_environment(self._profile.environments, os_environments)

        # migrate profile to latest schema
        self._profile.migrate()

        # merge profile inputs (add new / remove non-defined)
        self._profile.merge_inputs()

        # populate profile (env vars, etc)
        self._profile.data = self._profile.populate.replace_env_variables(self._profile.data)

        # validate input fields, this method requires env vars to be populated
        valid, message = self._profile.validate_inputs()

        # stage ThreatConnect data based on current profile, also used in teardown method
        self._staged_tc_data = self.stager.threatconnect.entities(
            self._profile.stage_threatconnect, self._profile.owner
        )
        # insert staged data for replacement
        self._profile.tc_staged_data = self._staged_tc_data

        # Replace staged_data
        self._profile.data = self._profile.populate.replace_tc_variables(self._profile.data)

        # replace all references and all staged variable
        self._profile.init()

        # stage kvstore data based on current profile
        self.stager.redis.from_dict(self._profile.stage_kvstore)

        return valid, message

    @property
    def profile(self):
        """Return profile instance."""
        return self._profile

    def run(self):
        """Implement in Child Class"""
        raise NotImplementedError('Child class must implement this method.')

    def run_app_method(self, app, method):
        """Run the provided App method."""
        try:
            getattr(app, method)()
        except SystemExit as e:
            self.log.data('run', 'Exit Code', e.code)
            if e.code != 0 and self.profile and e.code not in self.profile.exit_codes:
                self.log.data(
                    'run',
                    'App failed',
                    f'App exited with code of {e.code} in method {method}',
                    'error',
                )
            app.tcex.log.info(f'Exit Code: {e.code}')
            return e.code
        except Exception:
            self.log.data(
                'run',
                'App failed',
                f'App encountered except in {method}() method ({traceback.format_exc()})',
                'error',
            )
            return 1
        return 0

    @property
    def session_admin(self):
        """Return requests Session object for TC admin account.

        The credential this session uses require special activation in the ThreatConnect Platform
        and is not intended for normal use.
        """
        # APP-102 - adding a session for TC API Admin role request
        api_access_id = self.env_store.getenv('/ninja/tc/system/admin_api/api_access_id')
        api_secret_key = self.env_store.getenv('/ninja/tc/system/admin_api/api_secret_key')
        if (
            self._session_admin is None
            and api_access_id is not None
            and api_secret_key is not None
            and self.tc_api_path is not None
        ):
            # support for a proxy is not a typical use case, but can be added later if needed
            self._session_admin = TestSession(
                api_access_id=api_access_id,
                api_secret_key=api_secret_key,
                api_path=self.tc_api_path,
                logger=self.log,
            )
        return self._session_admin

    @property
    def session_exchange(self):
        """Return requests Session object for TC admin account.

        The credential this session uses require special activation in the ThreatConnect Platform
        and is not intended for normal use.
        """
        # APP-102 - adding a session for TC Exchange Admin role request
        if (
            self._session_exchange is None
            and self.api_access_id is not None
            and self.api_secret_key is not None
            and self.tc_api_path is not None
        ):
            # support for a proxy is not a typical use case, but can be added later if needed
            self._session_exchange = TestSession(
                api_access_id=self.api_access_id,
                api_secret_key=self.api_secret_key,
                api_path=self.tc_api_path,
                logger=self.log,
            )
        return self._session_exchange

    @classmethod
    def setup_class(cls):
        """Run once before all test cases."""
        cls.initialized = True
        cls._timer_class_start = time.time()
        cls.log.title('Setup Class', '#')
        TestCase.log.data('setup class', 'started', datetime.now().isoformat())
        TestCase.log.data('setup class', 'local envs', cls.env)

    def setup_method(self):
        """Run before each test method runs."""
        self._timer_method_start = time.time()
        self._current_test = os.getenv('PYTEST_CURRENT_TEST').split(' ')[0]
        self.log.title(self._current_test, '=')
        self.log.data('setup method', 'started', datetime.now().isoformat())

        # create and log current context
        self.context = os.getenv('TC_PLAYBOOK_DB_CONTEXT', str(uuid.uuid4()))
        self.log.data('setup method', 'context', self.context)

        # setup per method instance of tcex
        args = self.default_args.copy()
        args['tc_log_file'] = os.path.join(self.test_case_log_test_dir, 'setup.log')
        args['tc_logger_name'] = f'tcex-{self.test_case_feature}-{self.test_case_name}'
        self.tcex = TcEx(config=args)

        # initialize new stager instance
        self._stager = self.stager_init()

        # initialize new validator instance
        self._validator = self.validator_init()

        # Adding this for batch to created the -batch and errors files
        os.makedirs(os.path.join(self.test_case_log_test_dir, 'DEBUG'), exist_ok=True)

    @property
    def stager(self):
        """Return instance of Stager class."""
        return self._stager

    def stager_init(self):
        """Return instance of Stager class."""
        tc_log_file = os.path.join(self.test_case_log_test_dir, 'stage.log')

        # args data
        args = self.default_args.copy()

        # override default log level if profiled
        args['tc_log_level'] = 'warning'

        # set log path to be the feature and test case name
        args['tc_log_file'] = tc_log_file

        # set a logger name to have a logger specific for stager
        args['tc_logger_name'] = 'tcex-stager'

        tcex = TcEx(config=args)
        return Stager(tcex, logger)

    @property
    def tc_api_path(self):
        """Return TC Api Path."""
        return os.getenv('TC_API_PATH')

    @property
    def tc_token(self):
        """Return a valid API token.

        note:: requires TC >= 6.0
        """
        if self.tc_api_path is None:  # no API path, no token
            return None
        data = None
        token = None
        token_url_path = self.env_store.getenv('/ninja/tc/token/url_path', env_type='remote')
        if token_url_path is None:
            # could not retrieve URL path
            return None

        # determine the token type
        token_type = 'api'
        # per conversation with Marut, we should be able to just use api tokens
        # if self.ij.runtime_level.lower() in [
        #     'apiservice',
        #     'triggerservice',
        #     'webhooktriggerservice',
        # ]:
        #     data = {'serviceId': os.getenv('TC_TOKEN_SVC_ID', '441')}
        #     token_type = 'svc'

        # retrieve token from API using HMAC auth
        r = self.session_exchange.post(f'{token_url_path}/{token_type}', json=data, verify=True)
        if r.status_code == 200:
            token = r.json().get('data')
            self.log.data('setup', 'Using Token', token)
            self.log.data('setup', 'Token Elapsed', r.elapsed, 'trace')
        else:
            self.log.error(f'Failed to retrieve token ({r.text})')
        return token

    @classmethod
    def teardown_class(cls):
        """Run once before all test cases."""
        cls.initialized = False
        cls.log.title('Teardown Class', '^')
        TestCase.log.data('teardown class', 'finished', datetime.now().isoformat())
        TestCase.log.data('teardown class', 'elapsed', time.time() - cls._timer_class_start)

    def teardown_method(self):
        """Run after each test method runs."""
        if self.enable_update_profile and self.ij.runtime_level.lower() not in [
            'apiservice',
            'triggerservice',
            'webhooktriggerservice',
        ]:
            # exit message can not be validated for a Service App
            self.profile.update_exit_message()

        # delete threatconnect staged data
        self.stager.threatconnect.delete_staged(self._staged_tc_data)

        # log running times
        self.log.data('teardown method', 'finished', datetime.now().isoformat())
        self.log.data('teardown method', 'elapsed', time.time() - self._timer_class_start)

        # APP-262 - close out loggers
        self.tcex.logger.shutdown()
        self.stager.tcex.logger.shutdown()
        self.validator.tcex.logger.shutdown()

        # APP-262 - disconnect redis
        self.tcex.redis_client.connection_pool.disconnect()
        self.stager.tcex.redis_client.connection_pool.disconnect()
        self.validator.tcex.redis_client.connection_pool.disconnect()

        # update profile for session data
        if self.profile:  # doesn't exist for API services
            self.profile.session_manager.update_profile()

    @property
    def test_case_data(self):
        """Return partially parsed test case data."""
        return os.getenv('PYTEST_CURRENT_TEST').split(' ')[0].split('::')

    @property
    def test_case_feature(self):
        """Return partially parsed test case data."""
        return self.test_case_data[0].split('/')[1].replace('/', '-')

    @property
    def test_case_feature_dir(self):
        """Return profile fully qualified filename."""
        return os.path.join(self._app_path, 'tests', self.test_case_feature)

    @property
    def test_case_log_feature_dir(self):
        """Return profile fully qualified filename."""
        return os.path.join(self._app_path, os.getenv('TC_LOG_PATH', 'log'), self.test_case_feature)

    @property
    def test_case_log_test_dir(self):
        """Return profile fully qualified filename."""
        return os.path.join(self.test_case_log_feature_dir, self.test_case_name)

    @property
    def test_case_name(self):
        """Return partially parsed test case data."""
        return self.test_case_data[-1].replace('/', '-').replace('[', '-').replace(']', '')

    def validate_exit_message(self, test_exit_message, op='eq', **kwargs):
        """Validate App exit message."""
        if test_exit_message is not None:
            message_tc_file = os.path.join(
                os.getenv('TC_OUT_PATH', 'log'),
                self.test_case_feature,
                self.test_case_name,
                'message.tc',
            )
            app_exit_message = None
            if os.path.isfile(message_tc_file):
                with open(message_tc_file) as mh:
                    app_exit_message = mh.read()

                if app_exit_message:
                    kwargs['title'] = 'Exit Message Validation'
                    kwargs['log_app_data'] = json.dumps(app_exit_message)
                    if op == 'eq':
                        kwargs['log_test_data'] = json.dumps(test_exit_message)

                    # compare
                    passed, assert_error = self.validator.compare(
                        app_exit_message, test_exit_message, op=op, **kwargs
                    )
                    assert passed, assert_error
                else:
                    assert False, 'The message.tc file was empty.'
            else:
                assert False, f'No message.tc file found at ({message_tc_file}).'

    @property
    def validator(self):
        """Return instance of Stager class."""
        return self._validator

    def validator_init(self):
        """Return instance of Stager class."""
        tc_log_file = os.path.join(self.test_case_log_test_dir, 'validate.log')

        # args data
        args = self.default_args.copy()

        # override default log level if profiled
        args['tc_log_level'] = 'warning'

        # set log path to be the feature and test case name
        args['tc_log_file'] = tc_log_file

        # set a logger name to have a logger specific for stager
        args['tc_logger_name'] = 'tcex-validator'

        tcex = TcEx(config=args)
        return Validator(tcex, logger)



"""App Decorators Module."""
# third-party
import wrapt


class FailOnOutput:
    """Fail App if return value (output) value conditions are met.

    This decorator allows for the App to exit on conditions defined in the function
    parameters.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @FailOnOutput(
            fail_on=['false'], fail_msg='Operation returned a value of "false".'
        )
        def my_method(data):
            return data.lowercase()

    Args:
        fail_enabled (boolean|str, kwargs): Accepts a boolean or string value.  If a boolean value
            is provided that value will control enabling/disabling this feature. A string
            value should reference an item in the args namespace which resolves to a boolean.
            The value of this boolean will control enabling/disabling this feature.
        fail_msg (str, kwargs): The message to log when raising RuntimeError.
        fail_msg_property (str, kwargs): The App property containting the dynamic exit message.
        fail_on (list, kwargs): Defaults to None.
            Fail if return value from App method is in the list.
        write_output (bool, kwargs): Defaults to True.
            If true, will call App.write_outputs() before failing on matched fail_on value.
    """

    def __init__(self, **kwargs):
        """Initialize Class properties."""
        self.fail_enabled = kwargs.get('fail_enabled', True)
        self.fail_msg = kwargs.get('fail_msg', 'Method returned invalid output.')
        self.fail_on = kwargs.get('fail_on', [])
        self.fail_msg_property = kwargs.get('fail_msg_property')
        self.write_output = kwargs.get('write_output', True)

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def fail(app, *args, **kwargs):
            """Call the function and store or append return value.

            Args:
                app (class): The instance of the App class "self".
            """
            # call method to get output
            data = wrapped(*args, **kwargs)

            # self.enable (e.g., True or 'fail_on_false') enables/disables this feature
            enabled = self.fail_enabled
            if not isinstance(self.fail_enabled, bool):
                enabled = getattr(app.args, self.fail_enabled)
                if not isinstance(enabled, bool):  # pragma: no cover
                    raise RuntimeError(
                        'The fail_enabled value must be a boolean or resolved to bool.'
                    )
                app.tcex.log.debug(f'Fail enabled is {enabled} ({self.fail_enabled}).')

            failed = False
            if enabled:
                if isinstance(data, list):
                    # validate each value in the list of results.
                    for d in data:
                        if d in self.fail_on:
                            failed = True
                            break
                else:
                    if data in self.fail_on:
                        failed = True

                if failed:
                    if self.write_output:
                        app.tcex.playbook.write_output()
                        if hasattr(app, 'write_output'):
                            app.write_output()
                    app.exit_message = self.get_fail_msg(app)  # for test cases
                    app.tcex.exit(1, self.get_fail_msg(app))
            return data

        return fail(instance, *args, **kwargs)

    def get_fail_msg(self, app):
        """Return the appropriate fail message."""
        fail_msg = self.fail_msg
        if self.fail_msg_property and hasattr(app, self.fail_msg_property):
            fail_msg = getattr(app, self.fail_msg_property)
        return fail_msg



#!/usr/bin/env python
"""TcEx Framework Bin Command Base Module."""
# standard library
import os
import sys

# third-party
import colorama as c

from ..app_config_object import InstallJson, LayoutJson, TcexJson
from ..app_config_object.permutations import Permutations


class Bin:
    """Base Class for ThreatConnect command line tools.

    Args:
        _args (namespace): The argparser args Namespace.
    """

    def __init__(self, _args):
        """Initialize Class properties.

        Args:
            _args (namespace): The argparser args Namespace.
        """
        self.args = _args

        # properties
        self.app_path = os.getcwd()
        self.exit_code = 0
        self.ij = InstallJson()
        self.lj = LayoutJson()
        self.permutations = Permutations()
        self.tj = TcexJson()

        # initialize colorama
        c.init(autoreset=True, strip=False)

    @staticmethod
    def handle_error(err, halt=True):
        """Print errors message and optionally exit.

        Args:
            err (str): The error message to print.
            halt (bool, optional): Defaults to True. If True the script will exit.
        """
        print(f'{c.Style.BRIGHT}{c.Fore.RED}{err}')
        if halt:
            sys.exit(1)

    @staticmethod
    def print_message(message, line_bright=False, line_color=None, line_limit=150):
        """Print the message ensuring lines don't exceed line limit."""
        bright = ''
        if line_bright:
            bright = c.Style.BRIGHT
        message_line = ''
        for word in message.split(' '):
            if len(message_line) + len(word) < line_limit:
                message_line += f'{word} '
            else:
                print(f'{bright}{line_color}{message_line.rstrip()}')
                message_line = f'{word} '
        print(f'{bright}{line_color}{message_line.rstrip()}')

    @staticmethod
    def update_system_path():
        """Update the system path to ensure project modules and dependencies can be found."""
        cwd = os.getcwd()
        lib_dir = os.path.join(os.getcwd(), 'lib_')
        lib_latest = os.path.join(os.getcwd(), 'lib_latest')

        # insert the lib_latest directory into the system Path if no other lib directory found. This
        # entry will be bumped to index 1 after adding the current working directory.
        if not [p for p in sys.path if lib_dir in p]:
            sys.path.insert(0, lib_latest)

        # insert the current working directory into the system Path for the App, ensuring that it is
        # always the first entry in the list.
        try:
            sys.path.remove(cwd)
        except ValueError:
            pass
        sys.path.insert(0, cwd)



"""TcEx Webhook Trigger Server Test Case"""
# standard library
import sys

from .test_case_service_common import TestCaseServiceCommon


class TestCaseWebhookTriggerService(TestCaseServiceCommon):
    """WebhookTrigger App TestCase Class"""

    # TOOD: move to service command
    def run(self):
        """Run the Playbook App.

        Returns:
            int: The App exit code
        """
        # first-party
        from run import run  # pylint: disable=no-name-in-module

        # backup sys.argv
        sys_argv_orig = sys.argv

        # clear sys.argv
        sys.argv = sys.argv[:1]

        # run the app
        exit_code = 0
        try:
            run(set_app=self._app_callback)  # pylint: disable=unexpected-keyword-arg
        except SystemExit as e:
            exit_code = e.code

        # restore sys.argv
        sys.argv = sys_argv_orig

        self.log.data('run', 'Exit Code', exit_code)
        return exit_code



"""ThreatConnect TI Incident"""
from ..group import Group


class Incident(Group):
    """Unique API calls for Incident API Endpoints

    Valid status:
    + Closed
    + Containment Achieved
    + Deleted
    + Incident Reported
    + Open
    + New
    + Rejected
    + Restoration Achieved
    + Stalled

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        event_date (str, kwargs): The incident event date expression for this Group.
        name (str, kwargs): [Required for Create] The name for this Group.
        status (str, kwargs): The status for this Group.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties."""
        super().__init__(
            tcex, sub_type='Incident', api_entity='incident', api_branch='incidents', **kwargs
        )

    def event_date(self, event_date):
        """Update the event_date.

        Args:
            event_date: Converted to %Y-%m-%dT%H:%M:%SZ date format.

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        event_date = self._utils.datetime.format_datetime(
            event_date, date_format='%Y-%m-%dT%H:%M:%SZ'
        )
        self._data['eventDate'] = event_date
        request = {'eventDate': event_date}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def status(self, status):
        """Update  the incidents status

        Valid status:
        + Closed
        + Containment Achieved
        + Deleted
        + Incident Reported
        + Open
        + New
        + Rejected
        + Restoration Achieved
        + Stalled

        Args:
            status: Closed, Containment Achieved, Deleted, Incident Reported, Open, New, Rejected,
            Restoration Achieved, Stalled.

        Returns:

        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        self._data['status'] = status
        request = {'status': status}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)
